import numpy as np
import pandas as pd
from numba import jit, prange
from scipy.optimize import fmin_slsqp


@jit(nopython=True)
def transform_data(a, b, std_data):
    """
    Revolatilize standardized data

    Parameters
    ----------
    a : float
        Alpha.
    b : float
        Beta.
    std_data : ndarray
        Standardized model data (t by k).

    Returns
    -------
    ndarray
        The revolatilized data.
    """
    t, k = std_data.shape
    var = np.ones((t, k))
    for i in range(1, t):
        var[i] = (1 - a - b) + a * std_data[i - 1] ** 2 + b * var[i - 1]
    return std_data * np.sqrt(var)


def cdcc_filter(params, std_data):
    """
    Correlation filtration for the cDCC model.

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    std_data : ndarray
        Standardized model data (t by k).

    Returns
    -------
    ndarray
        The t by k by k conditional correlation matrix.
    """
    a, b = params
    t, k = std_data.shape
    t_data = transform_data(a, b, std_data)
    qbar = t_data.T @ t_data / t
    intercept = (1 - a - b) * qbar
    r_t = np.zeros((t, k, k))
    qt = intercept + a * qbar + b * qbar
    for i in range(1, t+1):
        qstar = np.sqrt(np.diag(qt)).reshape((1, k))
        rt = qt / (qstar.T * qstar)
        r_t[i - 1] = rt
        qt = intercept + a * t_data[i - 1:i].T @ t_data[i - 1:i] + b * qt
    return r_t


@jit(nopython=True)
def cdcc_loglik(params, std_data):
    """
    Likelihood for the cDCC model

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    std_data : ndarray
        Standardized model data (t by k).

    Returns
    -------
    float
        The negative of the log likelihood (excluding constant terms).
    """

    log_lik = 0.0
    a, b = params
    t, k = std_data.shape
    t_data = transform_data(a, b, std_data)
    qbar = t_data.T @ t_data / t
    intercept = (1 - a - b) * qbar
    qt = intercept + a * qbar + b * qbar
    for i in range(1, t+1):
        qstar = np.sqrt(np.diag(qt)).reshape((1, k))
        rt = qt / (qstar.T * qstar)
        _, ld = np.linalg.slogdet(rt)
        rt_inv = np.linalg.inv(rt)
        s = std_data[i - 1:i]
        log_lik += ld + (s @ rt_inv @ s.T)[0, 0]
        qt = intercept + a * t_data[i - 1:i].T @ t_data[i - 1:i] + b * qt
    return log_lik


@jit(nopython=True)
def cdcc_pair_likelihood(a, b, t_data, std_data, qbar, m, n):
    """
    cDCC likelihood for a single pair.

    Parameters
    ----------
    a : float
        Alpha.
    b : float
        Beta.
    std_data : ndarray
        The standardized model data (t by k).
    qbar : ndarray
        The long-run covariance of the revolatilized standardized data.
    m : int
        Index of first asset.
    n : int
        Index of second asset.

    Returns
    -------
    float
        The negative of the log likelihood (excluding constant terms).
    """
    t = std_data.shape[0]
    log_lik = 0.0
    scale = 1 - a - b
    intercept_mm = scale * qbar[m, m]
    intercept_mn = scale * qbar[m, n]
    intercept_nn = scale * qbar[n, n]
    qt_mm = intercept_mm + (a + b) * qbar[m, m]
    qt_mn = intercept_mn + (a + b) * qbar[m, n]
    qt_nn = intercept_nn + (a + b) * qbar[n, n]
    for i in range(1, t+1):
        rho = qt_mn / np.sqrt(qt_mm * qt_nn)
        det = 1 - rho * rho
        ld = np.log(det)
        s_m = std_data[i - 1, m]
        s_n = std_data[i - 1, n]
        log_lik += ld + (s_m * s_m + s_n * s_n - 2 * rho * s_m * s_n) / det

        t_m = t_data[i - 1, m]
        t_n = t_data[i - 1, n]

        qt_mm = intercept_mm + a * t_m * t_m + b * qt_mm
        qt_mn = intercept_mn + a * t_m * t_n + b * qt_mn
        qt_nn = intercept_nn + a * t_n * t_n + b * qt_nn
    return log_lik


@jit(nopython=True, parallel=True)
def cdcc_composite_likelihood(params, std_data, how="mcle"):
    """
    Composite likelihood for the cDCC model

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    std_data : ndarray
        Standardized model data (t by k).
    how : {"mcle", "mscle"}
        Either "mcle" for all pairs for "mscle" for adjacent pairs.

    Returns
    -------
    float
        The negative of the composite log likelihood (excluding constant
        terms).
    """
    log_lik = 0.0
    a, b = params
    t, k = std_data.shape
    t_data = transform_data(a, b, std_data)
    qbar = t_data.T @ t_data / t
    if how == "full":
        for m in range(k):
            for n in prange(m + 1, k):
                log_lik += cdcc_pair_likelihood(a, b, t_data, std_data, qbar, m, n)
    elif how == "mscle":
        for m in range(k):
            n = (m + 1) % k
            log_lik += cdcc_pair_likelihood(a, b, t_data, std_data, qbar, m, n)
        if k == 2:
            log_lik /= 2

    return log_lik


def f_ieqcons(params, *args):
    """Inequality constraints used in estimation"""
    a, b = params
    return np.array([a, b, 1.0 - a - b])


def cdcc(std_data, method="mcle", disp=2):
    """
    cDCC model parameter estimation

    Parameters
    ----------
    std_data : ndarray
        Standardized residuals that should have conditional volaitliy 1 and
        mean 0. Must be t by k.
    method : {"mcle", "mscle", "mle"}
        The method to use. "mcle" is composite likelihood using all pairs.
        "mscle" uses adjacent pairs only. "mle" is full (Q)MLE.
    disp : int
        Verbosity of output. 0 for silent, 1 for final and 2 for iterative.

    Returns
    -------
    params : ndarray
        The optimal parameters (alpha, beta)
    rt : ndarray
        The estimated conditional correlations (t, k, k)
    """

    a_candidates = [.01, .03, .05, .1]
    ab_candidates = [.9, .95, .975, .99]
    initial_val = {}
    if method == "mle":
        lik_fn = cdcc_loglik
    elif method == "mcle":
        lik_fn = cdcc_composite_likelihood
    elif method == "mscle":
        from functools import partial
        lik_fn = partial(cdcc_composite_likelihood, how="pairs")
    else:
        raise NotImplementedError

    for a in a_candidates:
        for ab in ab_candidates:
            b = ab - a
            params = np.array([a, b])
            initial_val[(a, b)] = lik_fn(params, std_data)
    iv = pd.Series(initial_val)
    sv = np.array(iv.idxmin())
    bounds = [(0.0, 1.0), (0.0, 1.0)]

    opt = fmin_slsqp(lik_fn, sv, f_ieqcons=f_ieqcons, bounds=bounds, args=(std_data,), disp=disp)
    rt = cdcc_filter(opt, std_data)
    return opt, rt
