import numpy as np
import pandas as pd
from numba import jit, prange
from scipy.optimize import fmin_slsqp


@jit(nopython=True)
def scalar_bekk_filter(params, data, target):
    """
    Covariance filtration for covariance targeting Scalar BEKK

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    data : ndarray
        Model data (t by k).
    target : ndarray
        The long-run covariance target (k by k).

    Returns
    -------
    ndarray
        The t by k by k conditional covariance matrix.
    """
    a, b = params
    t, k = data.shape
    h_t = np.zeros((t, k, k))
    intercept = (1 - a - b) * target
    ht = intercept + (a + b) * target
    for i in range(1, t + 1):
        h_t[i - 1] = ht
        ht = intercept + a * data[i - 1:i].T @ data[i - 1:i] + b * ht
    return h_t


@jit(nopython=True)
def scalar_bekk_likelihood(params, data, target):
    """
    Log likelihood for targeting scalar BEKK model

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    data : ndarray
        Model data (t by k).
    target : ndarray
        The long-run covariance target (k by k).

    Returns
    -------
    float
        The negative of the composite log likelihood (excluding constant
        terms).
    """

    log_lik = 0.0
    a, b = params
    t, k = data.shape
    intercept = (1 - a - b) * target
    ht = intercept + (a + b) * target
    for i in range(1, t + 1):
        _, ld = np.linalg.slogdet(ht)
        ht_inv = np.linalg.inv(ht)
        log_lik += ld + (data[i - 1:i] @ ht_inv @ data[i - 1:i].T)[0, 0]
        ht = intercept + a * data[i - 1:i].T @ data[i - 1:i] + b * ht
    return log_lik


@jit(nopython=True)
def scalar_bekk_pair_likelihood(a, b, data, target, m, n):
    """
    Targeting Scalar BEKK Likelihood for a single pair.

    Parameters
    ----------
    a : float
        Alpha.
    b : float
        Beta.
    data : ndarray
        The model data (t by k).
    target : ndarray
        The covariance target (k by k).
    m : int
        Index of first asset.
    n : int
        Index of second asset.

    Returns
    -------
    float
        The negative of the log likelihood (excluding constant terms).
    """
    t = data.shape[0]
    log_lik = 0.0
    scale = 1 - a - b
    intercept_mm = scale * target[m, m]
    intercept_mn = scale * target[m, n]
    intercept_nn = scale * target[n, n]
    ht_mm = intercept_mm + (a + b) * target[m, m]
    ht_mn = intercept_mn + (a + b) * target[m, n]
    ht_nn = intercept_nn + (a + b) * target[n, n]
    for i in range(1, t + 1):
        det = ht_mm * ht_nn - ht_mn * ht_mn
        ld = np.log(det)
        d_m = data[i - 1, m]
        d_n = data[i - 1, n]
        # r' ht^-1 h
        inner_prod = (ht_nn * d_m * d_m +
                      ht_mm * d_n * d_n -
                      2 * ht_mn * d_m * d_n) / det
        log_lik += ld + inner_prod

        ht_mm = intercept_mm + a * d_m * d_m + b * ht_mm
        ht_mn = intercept_mn + a * d_m * d_n + b * ht_mn
        ht_nn = intercept_nn + a * d_n * d_n + b * ht_nn
    return log_lik


@jit(nopython=True, parallel=True)
def scalar_bekk_composite_likelihood(params, data, target, how="mcle"):
    """
    Composite likelihood for targeting scalar BEKK model

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    data : ndarray
        Model data (t by k).
    target : ndarray
        The long-run covariance target (k by k).
    how : {"mcle", "mscle"}
        Either "mcle" for all pairs for "mscle" for adjacent pairs.

    Returns
    -------
    float
        The negative of the composite log likelihood (excluding constant
        terms).
    """
    log_lik = 0.0
    a, b = params
    t, k = data.shape
    if how == "mcle":
        for m in range(k):
            for n in prange(m + 1, k):
                log_lik += scalar_bekk_pair_likelihood(a, b, data, target, m, n)
    elif how == "mscle":
        for m in range(k):
            n = (m + 1) % k
            log_lik += scalar_bekk_pair_likelihood(a, b, data, target, m, n)
        if k == 2:
            log_lik /= 2
    return log_lik


def f_ieqcons(params, *args):
    """Inequality constraints used in estimation"""
    a, b = params
    return np.array([a, b, 1.0 - a - b])


def scalar_bekk(data, method="mcle", disp=2):
    """
    Covariance-targeted Scalar BEKK model parameter estimation

    Parameters
    ----------
    data : ndarray
        Residuals that should have conditional mean 0. Must be t by k.
    method : {"mcle", "mscle", "mle"}
        The method to use. "mcle" is composite likelihood using all pairs.
        "mscle" uses adjacent pairs only. "mle" is full (Q)MLE.
    disp : int
        Verbosity of output. 0 for silent, 1 for final and 2 for iterative.

    Returns
    -------
    params : ndarray
        The optimal parameters (alpha, beta)
    ht : ndarray
        The estimated conditional covariance (t, k, k)
    """
    a_candidates = [.01, .03, .05, .1]
    ab_candidates = [.9, .95, .975, .99]
    initial_val = {}
    if method == "mle":
        lik_fn = scalar_bekk_likelihood
    elif method == "mcle":
        lik_fn = scalar_bekk_composite_likelihood
    elif method == "mscle":
        from functools import partial
        lik_fn = partial(scalar_bekk_composite_likelihood, how="mscle")
    else:
        raise NotImplementedError

    target = data.T @ data / data.shape[0]
    for a in a_candidates:
        for ab in ab_candidates:
            b = ab - a
            params = np.array([a, b])
            initial_val[(a, b)] = lik_fn(params, data, target)
    iv = pd.Series(initial_val)
    sv = np.array(iv.idxmin())
    bounds = [(0.0, 1.0), (0.0, 1.0)]
    args = (data, target)
    opt = fmin_slsqp(lik_fn, sv, f_ieqcons=f_ieqcons, bounds=bounds, args=args, disp=disp)
    ht = scalar_bekk_filter(opt, data, target)

    return opt, ht


def matrix_sqrt(m):
    """Compute a symmetric matrix square root"""
    val, vec = np.linalg.eigh(m)
    out = vec @ np.diag(np.sqrt(val)) @ vec.T
    return (out + out.T) / 2


def scalar_bekk_simulate(params, t, n, burn, seed=1234):
    """
    Scalar BEKK Simulation

    Parameters
    ----------
    params : ndarray
        The two parameters (alpha, beta).
    t : int
        The time series dimension of the simulated data.
    n : int
        The cross-sectional dimension of the simulated data.
    burn : int
        Number of observations to discard. t + burn are simulated.
    seed : int, list[int]
        Seed to use in the random number generator.

    Returns
    -------
    data : ndarray
        Simulated data (t by k).
    sigma_bar : ndarray
        The true value of the long-run variance (k by k).
    """

    a, b = params
    rng = np.random.RandomState(seed=seed)
    e = rng.standard_normal((burn + t, n))
    r = np.zeros((burn + t, n))
    rho = 0.5 + rng.random_sample((n, 1))
    omega = 2 + rng.chisquare(10, size=(n)) / 5
    sigma_bar = rho @ rho.T + np.diag(omega)
    sigma = sigma_bar.copy()
    r[[0]] = e[[0], :] @ matrix_sqrt(sigma)
    for i in range(1, burn + t):
        sigma = (1 - a - b) * sigma_bar + a * (r[[i - 1]].T @ r[[i - 1]]) + b * sigma
        r[[i]] = e[[i], :] @ matrix_sqrt(sigma)

    return r[burn:], sigma_bar
